"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.emitJSON = exports.meta = exports.schema = void 0;
const onigasm_1 = require("onigasm");
const promises_1 = require("fs/promises");
const path_1 = require("path");
exports.schema = "https://raw.githubusercontent.com/martinring/tmlanguage/master/tmlanguage.json";
/**
 * Special scope that indicates a larger construct that doesn't get a single color.
 * Expanded to meta.<key>.<grammar name> during emit.
 */
exports.meta = Symbol("meta");
let initialized = false;
async function initialize() {
    if (!initialized) {
        const path = path_1.resolve(__dirname, "../node_modules/onigasm/lib/onigasm.wasm");
        const wasm = await promises_1.readFile(path);
        await onigasm_1.loadWASM(wasm.buffer);
        initialized = true;
    }
}
/**
 * Emit the given grammar to JSON.
 */
async function emitJSON(grammar) {
    await initialize();
    const indent = 2;
    const processed = processGrammar(grammar);
    return JSON.stringify(processed, undefined, indent);
}
exports.emitJSON = emitJSON;
/**
 * Convert the grammar from our more convenient representation to the
 * tmlanguage.json schema. Perform some validation in the process.
 */
function processGrammar(grammar) {
    // key is rule.key, value is [unprocessed rule, processed rule]. unprocessed
    // rule is used for its identity to check for duplicates and deal with cycles.
    const repository = new Map();
    const output = processNode(grammar);
    output.repository = processRepository();
    return output;
    function processNode(node) {
        if (typeof node !== "object") {
            return node;
        }
        if (Array.isArray(node)) {
            return node.map(processNode);
        }
        const output = {};
        for (const key in node) {
            const value = node[key];
            switch (key) {
                case "key":
                    // Drop it. It was used to place the node in the repository, and does
                    // not need to be retained on the node in the final structure.
                    break;
                case "scope":
                    // tmlanguage uses "name" confusingly for scope. We avoid "name" which
                    // can be confused with the repository key.
                    output.name = value === exports.meta ? `meta.${node.key}.adl` : value;
                    break;
                case "begin":
                case "end":
                case "match":
                    validateRegexp(value, node, key);
                    output[key] = value;
                    break;
                case "patterns":
                    output[key] = processPatterns(value);
                    break;
                default:
                    output[key] = processNode(value);
                    break;
            }
        }
        return output;
    }
    function processPatterns(rules) {
        for (const rule of rules) {
            if (!repository.has(rule.key)) {
                // put placeholder first to prevent cycles
                const entry = [rule, undefined];
                repository.set(rule.key, entry);
                // fill placeholder with processed node.
                entry[1] = processNode(rule);
            }
            else if (repository.get(rule.key)[0] !== rule) {
                throw new Error("Duplicate key: " + rule.key);
            }
        }
        return rules.map((r) => ({ include: `#${r.key}` }));
    }
    function processRepository() {
        const output = {};
        for (const key of [...repository.keys()].sort()) {
            output[key] = repository.get(key)[1];
        }
        return output;
    }
    function validateRegexp(regexp, node, prop) {
        try {
            new onigasm_1.OnigRegExp(regexp).testSync("");
        }
        catch (err) {
            if (/^[0-9,]+/.test(err.message)) {
                // Work around for https://github.com/NeekSandhu/onigasm/issues/26
                const array = new Uint8Array(err.message.split(",").map((s) => Number(s)));
                const buffer = Buffer.from(array);
                err = new Error(buffer.toString("utf-8"));
            }
            console.error(`Error: Bad regex: ${JSON.stringify({ [prop]: regexp })} in:`);
            console.error(node);
            throw err;
        }
    }
}
//# sourceMappingURL=index.js.map